module UU
  module OS
    module Util

      # Module providing methods for +snake_case+ to +camelCase+ (and vice versa) conversion.
      # @private
      module NameConverter # :nodoc:

        # Converts given name to +snake_case+. Works only on names containing
        # alphanumeric characters (excluding national characters) with optional
        # underscore prefix and postfix.
        # @param [String, Symbol] name Name in +camelCase+
        # @return [String, Symbol] Name in +snake_case+
        def to_snake_case(name)
          NameConverter.to_snake_case(name)
        end

        # Converts given name to +lowerCamelCase+ (where each word with exception of first
        # one starts with upper case followed by lower cases). Works only on names containing
        # alphanumeric characters (excluding national characters) with optional underscore
        # prefix and postfix.
        # @param [String, Symbol] name Name in +snake_case+
        # @return [String, Symbol] Name in +camelCase+
        def to_camel_case(name)
          NameConverter.to_camel_case(name)
        end

        # Converts given name to +UpperCamelCase+ (where each word including of first one
        # starts with upper case followed by lower cases). Works only on names containing
        # alphanumeric characters (excluding national characters) with optional underscore
        # prefix and postfix.
        # @param [String, Symbol] name Name in +snake_case+
        # @return [String, Symbol] Name in +UpperCamelCase+
        def to_upper_camel_case(name)
          NameConverter.to_upper_camel_case(name)
        end

        # @see #to_snake_case
        # @private
        def self.to_snake_case(name) # :nodoc:
          return name if name.nil? || name.empty?
          raise ArgumentError, "\"#{name}\" cannot be reliably converted to snake_case" if name !~ /^[_]*[\$]?[a-zA-Z0-9_]+$/
          # Separate prefix and postfix
          name =~ /(^[_]*[\$]?)/
          prefix = $1
          name =~ /([_]*$)/
          postfix = $1
          # In case name contain underscores, leave them unchanged
          elements = name.to_s[prefix.size..(-1 - postfix.size)].split('_')
          selements = []
          elements.each do |element|
            # Tokenize element to uppercase, lowercase and numeric groups
            tokens = []
            element.gsub(/([a-z]+|[A-Z]+|[0-9]+)/) { |e| tokens << $1}
            # Concatenate tokens to valid snake_case name
            sname = ''
            tokens.each_with_index do |token, idx|
              if token =~ /^[a-z]+$/
                # Lowercase group is not separator.
                sname << token
              else
                # Append separator if there is none and we are not first element.
                sname << '_' if idx > 0 && sname !~ /_$/
                if token.size == 1 && token !~ /^[0-9]+$/
                  # Single capital letter is part of next lowercase token
                  sname << token
                else
                  if token =~ /^[0-9]+$/ || idx == tokens.size - 1 || tokens[idx + 1] =~ /^[0-9]+$/
                    # Number, last token and token followed by numeric token must be added whole
                    sname << token
                    # Append separator if there is more tokens
                    sname << '_' if idx < tokens.size - 1
                  else
                    # Group of uppercase letters must be split so that last
                    # letter is part of next token
                    sname << token[0..-2]
                    sname << '_'
                    sname << token[-1]
                  end
                end
              end
            end
            selements << sname
          end
          # Join prefix, name (which may contain multiple parts) and postfix, and return as lowercase
          result = "#{prefix}#{selements.join('_')}#{postfix}".downcase
          result = result.to_sym if name.kind_of?(Symbol)
          return result
        end

        # @see #to_camel_case
        # @private
        def self.to_camel_case(name) # :nodoc:
          return name if name.nil? || name.empty?
          raise ArgumentError, "\"#{name}\" cannot be reliably converted to camelCase" if name !~ /^[_]*[\$]?[a-zA-Z0-9_]+$/
          # First convert to snake case to unify style of abbreviations (all upper case of only first letter uppercase)
          name = to_snake_case(name)
          # Separate prefix and postfix
          name =~ /(^[_]*[\$]?)/
          prefix = $1
          name =~ /([_]*$)/
          postfix = $1
          # Remove underscores and capitalize each letter next to underscore
          cname = name.to_s[prefix.size..(-1 - postfix.size)].gsub(/[_]+(.)/) { |e| $1.upcase}
          # Join prefix, name and postfix and return
          result = "#{prefix}#{cname}#{postfix}"
          result = result.to_sym if name.kind_of?(Symbol)
          return result
        end

        # @see #to_camel_case
        # @private
        def self.to_upper_camel_case(name) # :nodoc:
          return name if name.nil? || name.empty?
          raise ArgumentError, "\"#{name}\" cannot be reliably converted to UpperCamelCase" if name !~ /^[_]*[\$]?[a-zA-Z0-9_]+$/
          result = to_camel_case(name).to_s
          result[0] = result[0].upcase
          result = result.to_sym if name.kind_of?(Symbol)
          return result
        end

      end

    end
  end
end
